/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include "PartIdSet.h"
#include <Database.h>
#include <Type.h>
#include <Reference.h>
#include <GlobalEntities.h>
#include <Group.h>

extern Group* oo7Group;

//--------------------------------------------------------------------------
// Start with some necessary preliminaries
//--------------------------------------------------------------------------

#define TypeSize	10
#define TitleSize	40


typedef enum {Complex, Base} AssemblyType;

typedef enum {Trav1, Trav1WW, Trav2a, Trav2b, Trav2c, Trav3a, Trav3b, 
	Trav3c, Trav4, Trav5do, Trav5undo, Trav6, Trav7, Trav8, Trav9, 
	Trav10,
	Query1, Query2, Query3, Query4, Query5, Query6, Query7, Query8, 
	Insert, Delete, Reorg1, Reorg2,
	WarmUpdate, MultiTrav1, MultiTrav2, MultiTrav3, MultiTrav4, 
	MultiTrav5, MultiTrav6} BenchmarkOp;

typedef enum {UpdateOne, UpdateAll, UpdateRepeat} UpdateType;
typedef enum {UpdateDirectionDo, UpdateDirectionUndo} UpdateDirectionType;

extern class Assembly;
extern class BaseAssembly;
extern class ComplexAssembly;
extern class CompositePart;
extern class Document;
extern class Module;

//--------------------------------------------------------------------------
// AtomicPart objects are the primitives for building up designs
//	- modeled after the Sun/OO1 benchmark's parts
//--------------------------------------------------------------------------

class AtomicPart : public Object {
private:
    Reference   compPartReference;
    Reference   toListReference;
    Reference   fromListReference;
public:
    AtomicPart(int ptId, CompositePart*  cp);
    ~AtomicPart();

    // magic required by Ontos.
    AtomicPart(APL* theAPL);

    // required by Ontos.
    Type *getDirectType();
    void Destroy (OC_Boolean = FALSE);
    void deleteObject (OC_Boolean = TRUE);

    int	    id;
    char    type[TypeSize];
    int	    buildDate;
    int	    x, y;
    int     docId;

    List*    to() {    // to connection objects
        return (List*) toListReference.Binding(this);
    };

    List*    from() {    // from connection objects
        return (List*) fromListReference.Binding(this);
    };

    CompositePart*  partOf() { // create up pointer from Reference
       return (CompositePart*) compPartReference.Binding(this);
    };

    void swapXY();
    void toggleDate();
    void DoNothing();

    int traverse(BenchmarkOp op, PartIdSet& visitedIds);
};

//--------------------------------------------------------------------------
// Connection objects are used to wire AtomicParts together
//	- similarly, modeled after Sun/OO1 connections
//--------------------------------------------------------------------------

class Connection : public Object {
private:
    Reference fromReference;
    Reference toReference;

public:
    char		type[TypeSize];
    int		        length;

    class AtomicPart*	from() {  // create memory pointer from Reference
       return (AtomicPart*) fromReference.Binding(this);
    };

    class AtomicPart*	to() {  // create memory pointer from Reference
       return (AtomicPart*) toReference.Binding(this);
    };

    Connection(AtomicPart* fromPart, AtomicPart* toPart);
    ~Connection(){};

    // magic required by Ontos.
    Connection(APL*);

    // required by Ontos.
    Type *getDirectType();
};


//--------------------------------------------------------------------------
// CompositeParts are parts constructed from AtomicParts
//	- entry in a library of reusesable components
//	- implementation is a graph of atomic parts
//	- provides unit of significant access locality
//	- each has an associated (unique) document object
//--------------------------------------------------------------------------

class CompositePart : public Object {
private:
    Reference           documentReference;
    Reference           rootPartReference;
    Reference           partsReference;
    Reference           privReference;
    Reference           sharReference;

public:
    int		id;
    char	type[TypeSize];
    int		buildDate;

    Document*	documentation() { // create pointer from Reference
        return (Document*) documentReference.Binding(storageManager());
    };

    AtomicPart*	rootPart() { // create pointer from Reference
        return (AtomicPart*) rootPartReference.Binding(this);
    }

    // all atomic parts in the composite part (useful in delete.)
    Set* parts() {
        return (Set*) partsReference.Binding(this);
    }

    // all base assemblies that have private references to this part 
    List* usedInPriv() {
        return (List*) privReference.Binding(this);
    }

    // all base assemblies that have shared references to this part 
    List* usedInShar() {
        return (List*) sharReference.Binding(this);
    }


    CompositePart(int cpId);

    // magic constructor required by Ontos.
    CompositePart(APL *theAPL);

    Type *getDirectType();
    void Destroy (OC_Boolean = FALSE);
    void deleteObject (OC_Boolean = TRUE);

    ~CompositePart();
    int traverse(BenchmarkOp op);
    int traverse7();
    int reorg1();
    int reorg2();
};


//--------------------------------------------------------------------------
// Document objects are used to hold a description of some particular
// CompositePart object
//--------------------------------------------------------------------------

class Document : public Object {
private:
    Reference                   compPartReference;
public:
    char			title[TitleSize];
    int                         id;
    char*			text;
    int	 		        textSize; // actual size of text area
    int			        textLen;  // amount actually being used

    CompositePart*	        part() {    // get C pointer from reference
        return (CompositePart*) compPartReference.Binding(this);
    }
    /*
    was
        return (CompositePart*) compPartReference.Binding(oo7Group);
*/

    Document(int cpId, CompositePart* part);
    ~Document();

    // magic required by Ontos.
    Document(APL*);

    // required by Ontos.
    Type *getDirectType();

    int searchText(char c);
    int replaceText(char *oldString, char* newString);
};

//--------------------------------------------------------------------------
// Manual objects are used to hold a description of some particular
// module.  They are really just Documents for Modules, but it is
// cleaner to make them a different type, especially in the systems
// that automatically maintain extents.
//--------------------------------------------------------------------------

class Manual : public Object {
private:
    Reference                   moduleReference;
public:
    char			title[TitleSize];
    int                         id;
    char*			text;
    int	 		        textSize; // actual size of text area
    int			        textLen;  // amount actually being used

    Module*	        mod() {    // get C pointer from reference
        return (Module*) moduleReference.Binding(this);
    }

    Manual(int modId, Module* modPtr);
    ~Manual();

    // magic required by Ontos.
    Manual(APL*);

    // required by Ontos.
    Type *getDirectType();

    int searchText(char c);
    int replaceText(char *oldString, char* newString);
    int firstLast();
};

//--------------------------------------------------------------------------
// Assembly objects are design instances built up recursively from
// from other Assembly objects and (at the leaves only) CompositeParts
//	- hierarchical (tree) structure for designs
//	- may share composite parts with other assemblies
//	- nonleaf and leaf assembly subtypes
//--------------------------------------------------------------------------

class Assembly : public Object {
public:
    Reference assemblyReference;
    int			        id;
    char			type[TypeSize];
    int			        buildDate;
    AssemblyType		assmType;
    ComplexAssembly*	        superAssembly() { // return pointer
        return (ComplexAssembly*) assemblyReference.Binding(this);
    } 

    // magic required by Ontos.
    Assembly(APL* theAPL);

    // required?  Ontos seems to need it.
    Assembly() {};

    // required by Ontos.
    Type *getDirectType();

    virtual int traverse(BenchmarkOp op) = 0;
    virtual int traverse7(PartIdSet& visitedComplexIds) = 0;
    virtual AssemblyType mytype() = 0;
    void DoNothing();
};


class ComplexAssembly: public Assembly {
    Reference   subAssembliesReference;
public:

    List*	subAssemblies() {  // return pointer to list
        return (List*) subAssembliesReference.Binding(this);
    }

    ComplexAssembly(int asId, Module* module, 
	ComplexAssembly* parentAssembly, int levelNo);

    // magic required by Ontos.
    ComplexAssembly(APL*);
    ~ComplexAssembly();

    // required by Ontos.
    Type *getDirectType();
    void Destroy (OC_Boolean = FALSE);
    void deleteObject (OC_Boolean = TRUE);

    int traverse(BenchmarkOp op);
    int traverse7(PartIdSet& visitedComplexIds);
    virtual AssemblyType mytype() { return Complex;}
};


class BaseAssembly: public Assembly {
public:
    Reference privReference;    // for componentsPriv list
    Reference sharReference;    // for componentsShar list

    List*	componentsPriv() { // list of private composite parts
        return (List*) privReference.Binding(this);
    }

    List*	componentsShar() { // list of shared composite parts
        return (List*) sharReference.Binding(this);
    }


    BaseAssembly(int asId, Module* module,
    	             ComplexAssembly* parentAssembly);

    // Required by ONTOS.
    BaseAssembly(APL* theAPL);
    ~BaseAssembly();

    // required by Ontos.
    Type *getDirectType();
    void Destroy (OC_Boolean = FALSE);
    void deleteObject (OC_Boolean = TRUE);

    int traverse(BenchmarkOp op);
    int traverse7(PartIdSet& visitedComplexIds);
    virtual AssemblyType mytype() { return Base;}
};


//--------------------------------------------------------------------------
// Modules are the designs resulting from Assembly composition
//	- unit of scaleup for the benchmark database
//	- may share composite parts with other modules
//--------------------------------------------------------------------------

class Module : public Object {
private:
    Reference           designRootReference;
    Reference           allBasesReference;
    Reference           manualReference;

public:
    int			id;
    char		type[TypeSize];
    int			buildDate;

    ComplexAssembly*    designRoot() { 
        return (ComplexAssembly*) designRootReference.Binding(this);
    };

    Set*    allBases() { 
        return (Set*) allBasesReference.Binding(this);
    };

    Manual*    man() { 
        return (Manual*) manualReference.Binding(this);
    };


    Module(int modId);
  
    ~Module();

    // magic required by Ontos.
    Module(APL*);

    // required by Ontos.
    Type *getDirectType();
    void Destroy (OC_Boolean = FALSE);
    void deleteObject (OC_Boolean = TRUE);


    int traverse(BenchmarkOp op);

    int scanManual();

    int firstLast();
};


//--------------------------------------------------------------------------
// Dummy objects are used to flush the cache.
//--------------------------------------------------------------------------

class Dummy : public Object {
public:
    char*			text;

    Dummy();
    ~Dummy();

    // magic required by Ontos.
    Dummy(APL*);

    // required by Ontos.
    Type *getDirectType();

};

